module Feature.Sharing.SQLite where

import           Core.SQLite
import           Core.Types
import           Feature.Generation.Passwords (Password)

import           Control.Monad.Reader         (ask, lift, liftIO)
import           Crypto.Simple.CBC            (decrypt, encrypt)
import           Data.List.Split              (splitOn)
import           Data.Maybe                   (listToMaybe)
import           Data.Time.Clock.POSIX        (getPOSIXTime)
import           Database.SQLite.Simple

import qualified Data.ByteString.Base64       as B64
import qualified Data.ByteString.Char8        as B
import qualified Data.Text                    as T
import qualified Data.Text.Encoding           as ET
import qualified Data.Text.Lazy               as LT

findByLink :: String -> PurrAction (Maybe T.Text)
findByLink link = do
  db    <- dbPath
  key   <- encKey
  conn  <- liftIO $ open db
  res   <- liftIO $ query conn "SELECT * from pws WHERE link = ?" (Only (last $ splitOn "/" link))
  liftIO $ close conn
  readEncryptedSecret key res

insertNewSecret :: T.Text -> Integer -> T.Text -> Integer -> PurrAction ()
insertNewSecret sec life link maxViews = do
  db     <- dbPath
  key    <- encKey
  encSec <- liftIO $ encryptSecret key sec
  conn   <- liftIO $ open db
  time   <- liftIO $ epochTime
  liftIO  $ execute conn
    "INSERT INTO pws (link, secret, date, life, views, maxViews) VALUES (?, ?, ?, ?, ?, ?)"
      (SecretEntry link (encodeSecret encSec) time life 0 maxViews)
  liftIO  $ close conn

readEncryptedSecret :: String -> [SecretEntry] -> PurrAction (Maybe T.Text)
readEncryptedSecret key sec = do
  db     <- dbPath
  liftIO  $ incViews sec db
  delete <- liftIO $ deleteExpiredSecret sec db
  decKey <- liftIO ( sequence
                   $ decryptSecret key
                 <$> decodeSecret
                 <$> listToMaybe sec )
  if (delete)
    then return Nothing
    else return (ET.decodeLatin1 <$> decKey)
    where
      incViews :: [SecretEntry] -> String -> IO ()
      incViews []         _ = return ()
      incViews (sec : _) db = do
        conn <- liftIO $ open db
        liftIO $ execute conn
          "UPDATE pws SET views = views + 1 WHERE link = ?" (Only (link sec))
        liftIO $ close conn

-- Returns True if deletion occurs, informing the caller to not provide
-- the successfully retrieved secret to the requestor.
deleteExpiredSecret :: [SecretEntry] -> String -> IO Bool
deleteExpiredSecret []         _ = return False
deleteExpiredSecret (sec : _) db = do
  time <- liftIO $ epochTime
  if ((date sec) + ((life sec) * 86400) < time) || (views sec >= maxViews sec)
    then deleteSec sec db
    else return False
    where
      deleteSec :: SecretEntry -> String -> IO Bool
      deleteSec sec db = do
        conn  <- liftIO $ open db
        liftIO $ execute conn
          "DELETE FROM pws WHERE link = ?" (Only (link sec))
        liftIO $ close conn
        return True

encodeSecret :: B.ByteString -> T.Text
encodeSecret b = ET.decodeUtf8 $ B64.encode b

decodeSecret :: SecretEntry -> B.ByteString
decodeSecret s = B64.decodeLenient $ ET.encodeUtf8 (secret s)

encryptSecret :: String -> T.Text -> IO B.ByteString
encryptSecret k s = encrypt (B.pack k) (ET.encodeUtf8 s)

decryptSecret :: String -> B.ByteString -> IO B.ByteString
decryptSecret k b = decrypt (B.pack k) b

epochTime :: IO Integer
epochTime = fmap round getPOSIXTime
